@@ -33,26 +33,307 @@ from mlstudio.utils.validation import validate_objective, validate_optimizer
 from mlstudio.utils.validation import validate_scorer, validate_string
 from mlstudio.utils.validation import validate_zero_to_one, validate_int
 from mlstudio.utils.validation import validate_metric
+# --------------------------------------------------------------------------  #
+#                      LEARNING RATE SCHEDULE HELPERS                         #
+# --------------------------------------------------------------------------  #
+def compute_total_batches(estimator):
+    """Computes the total number of batches to be processed.
+
+    Parameters
+    ----------
+    estimator : MLStudio estimator object.
+    
+    """
+    m = estimator.X_train_.shape[0]
+    batch_size = estimator.batch_size or m
+    return estimator.epochs * math.ceil(m / batch_size)
+# --------------------------------------------------------------------------  #
+def compute_default_decay_factor(estimator):
+    """Computes a default decay factor
+
+     Parameters
+    ----------
+    estimator : MLStudio estimator object.
+    
+    """
+    return estimator.learning_rate / compute_total_batches(estimator)
 
 # --------------------------------------------------------------------------  #
-class LearningRateSchedule(Callback):
+#                      LEARNING RATE SCHEDULE FUNCTIONS                       #
+# --------------------------------------------------------------------------  #
+# Functions that perform the learning rate calculation in the callback. This 
+# functionality has been pulled from the callback to facilitate unit testing.
+# --------------------------------------------------------------------------  #
+def constant(attributes, eta0, batch, **kwargs):
+    """ Constant learning rate
+
+    Parameters
+    ----------
+    attributes : dict
+        Dictionary containing the attributes of the LearningRateSchedule object
+
+    eta0 : float
+        The initial learning rate
+
+    batch : int
+        The batch number. For batch gradient descent, this is the same
+        as the epoch number.
+
+    kwargs : dict
+        There are no additional keyword arguments for this function.   
+
+    """    
+    return eta0
+# --------------------------------------------------------------------------  #
+def step_decay(attributes, eta0, batch, **kwargs):
+    """ Step decay learning rate schedule as:
+    .. math:: \eta_0 \times \gamma^{\text{floor((1+batch)/decay_steps)}}
+
+    Parameters
+    ----------
+    attributes : dict
+        Dictionary containing the attributes of the LearningRateSchedule object
+
+    eta0 : float
+        The initial learning rate
+
+    batch : int
+        The batch number. For batch gradient descent, this is the same
+        as the epoch number.
+
+    kwargs : dict
+        There are no additional keyword arguments for this function.   
+
+    """    
+    decay_factor = attributes.get('decay_factor')
+    decay_steps = attributes.get('decay_steps')
+    assert decay_factor, "step_decay function requires a decay_factor attribute."
+    assert decay_steps, "step_decay function requires a decay_steps attribute."
+    return eta0 * np.power(decay_factor, math.floor((1+batch)/decay_steps))
+
+# --------------------------------------------------------------------------  #
+def time_decay(attributes, eta0, batch, **kwargs):
+    """ Time decay learning rate schedule as:
+    .. math:: \eta_t=\frac{\eta_0}{1+b\cdot t} 
+
+    Parameters
+    ----------
+    attributes : dict
+        Dictionary containing the attributes of the LearningRateSchedule object
+
+    eta0 : float
+        The initial learning rate
+
+    batch : int
+        The batch number. For batch gradient descent, this is the same
+        as the epoch number.
+
+    kwargs : dict
+        There are no additional keyword arguments for this function.        
+        
+    """
+    decay_factor = attributes.get('decay_factor')    
+    assert decay_factor, "time_decay function requires a decay_factor attribute."    
+    return eta0 / (1 + decay_factor * batch)   
+
+# --------------------------------------------------------------------------  #
+def sqrt_time_decay(attributes, eta0, batch, **kwargs):
+    """ Square root time decay learning rate schedule as:
+    .. math:: \eta_t=\frac{\eta_0}{1+b\cdot \sqrt{t}} 
+
+    Parameters
+    ----------
+    attributes : dict
+        Dictionary containing the attributes of the LearningRateSchedule object
+
+    eta0 : float
+        The initial learning rate
+
+    batch : int
+        The batch number. For batch gradient descent, this is the same
+        as the epoch number.
+
+    kwargs : dict
+        There are no additional keyword arguments for this function.     
+
+    """
+    decay_factor = attributes.get('decay_factor')    
+    assert decay_factor, "sqrt_time_decay function requires a decay_factor attribute."    
+    return eta0 / (1 + decay_factor * np.sqrt(batch))      
+
+
+# --------------------------------------------------------------------------  #
+def exponential_decay(attributes, eta0, batch, **kwargs):
+    """ Exponential decay learning rate schedule as:
+    .. math:: \eta_t=\eta_0 \cdot \text{exp}(-b\cdot t)
+
+    Parameters
+    ----------
+    attributes : dict
+        Dictionary containing the attributes of the LearningRateSchedule object
+
+    eta0 : float
+        The initial learning rate
+
+    batch : int
+        The batch number. For batch gradient descent, this is the same
+        as the epoch number.
+
+    kwargs : dict
+        There are no additional keyword arguments for this function.       
+        
+    """
+    decay_factor = attributes.get('decay_factor')    
+    assert decay_factor, "exponential_decay function requires a decay_factor attribute."    
+    return eta0  * np.exp(-decay_factor * batch)    
+
+# --------------------------------------------------------------------------  #
+def polynomial_decay(attributes, eta0, batch, **kwargs):
+    """ Exponential decay learning rate schedule as:
+    .. math:: \eta_t=\eta_0 \cdot \text{exp}(-b\cdot t)
+
+    Parameters
+    ----------
+    attributes : dict
+        Dictionary containing the attributes of the LearningRateSchedule object
+
+    eta0 : float
+        The initial learning rate
+
+    batch : int
+        The batch number. For batch gradient descent, this is the same
+        as the epoch number.
+
+    kwargs : dict
+        total_batches : int
+            The total number of anticipated batches to be processed.   
+        
+    """
+    power = attributes.get('power')    
+    total_batches = kwargs.get('total_batches')
+    assert power, "polynomial_decay function requires a power attribute."    
+    assert total_batches, "total_batches not sent to polynomial_decay function."
+    decay = (1-(batch / float(total_batches))) ** power
+    return eta0  * decay
+# --------------------------------------------------------------------------  #
+def exponential_step_decay(attributes, eta0, batch, **kwargs):
+    """ Exponential step decay learning rate schedule as:
+    .. math:: \eta_t=\eta_0 \cdot 10^{\frac{-t}{r}}
+
+    Parameters
+    ----------
+    attributes : dict
+        Dictionary containing the attributes of the LearningRateSchedule object
+
+    eta0 : float
+        The initial learning rate
+
+    batch : int
+        The batch number. For batch gradient descent, this is the same
+        as the epoch number.
+
+    kwargs : dict
+        There are no additional keyword arguments for this function.       
+
+    """
+    decay_factor = attributes.get('decay_factor')
+    decay_steps = attributes.get('decay_steps')
+    assert decay_factor, "decay_factor not sent to exponential_step_decay function."
+    assert decay_steps, "decay_steps not sent to exponential_step_decay function."
+    return eta0 * np.power(decay_factor, (batch / decay_steps))
+
+# --------------------------------------------------------------------------  #
+def power_schedule(attributes, eta0, batch, **kwargs):
+    """ Power learning rate schedule as:
+    .. math:: \eta_t=\eta_0 / (1+\frac{t}{r})^{c}
+    
+    Parameters
+    ----------
+    attributes : dict
+        Dictionary containing the attributes of the LearningRateSchedule object
+
+    eta0 : float
+        The initial learning rate
+
+    batch : int
+        The batch number. For batch gradient descent, this is the same
+        as the epoch number.
+
+    kwargs : dict
+        There are no additional keyword arguments for this function.       
+
+    """
+    power = attributes.get('power')
+    decay_steps = attributes.get('decay_steps')
+    assert power, "power not sent to power_schedule function."
+    assert decay_steps, "decay_steps not sent to power_schedule function."
+    return eta0 / (1 + batch/decay_steps)**power
+
+# --------------------------------------------------------------------------  #
+def bottou_schedule(attributes, eta0, batch, **kwargs):
+    """ Learning rate schedule as described in:
+    https://cilvr.cs.nyu.edu/diglib/lsml/bottou-sgd-tricks-2012.pdf
+    
+    Parameters
+    ----------
+    attributes : dict
+        Dictionary containing the attributes of the LearningRateSchedule object
+
+    eta0 : float
+        The initial learning rate
+
+    batch : int
+        The batch number. For batch gradient descent, this is the same
+        as the epoch number.
+
+    kwargs : dict
+        There are no additional keyword arguments for this function.       
+
+    """
+    decay_factor = attributes.get('decay_factor')    
+    assert decay_factor, "decay_factor not sent to bottou_schedule function."
+    return eta0 * (1 + eta0 * decay_factor * batch)**(-1)
+
+# --------------------------------------------------------------------------  #
+#                      LEARNING RATE SCHEDULE FACTORY                         #
+# --------------------------------------------------------------------------  #
+def get_learning_rate_function(schedule):
+    """Returns the learning rate function for a given learning rate schedule class."""
+    functions = {'Constant': constant, 'StepDecay': step_decay, 
+                 'TimeDecay': time_decay, 
+                 'SqrtTimeDecay': sqrt_time_decay, 
+                 'ExponentialDecay': exponential_decay,
+                 'PolynomialDecay': polynomial_decay,
+                 'ExponentialStepDecay': exponential_step_decay,
+                 'PowerSchedule': power_schedule,
+                 'BottouSchedule': bottou_schedule}
+
+    classname = schedule.__class__.__name__
+    return functions[classname]
+
+
+# --------------------------------------------------------------------------  #
+#                      LEARNING RATE SCHEDULE CALLBACKS                       #
+# --------------------------------------------------------------------------  #
+class LearningRateSchedule(ABC, Callback):
     """Base class for learning rate schedules. """
 
     @abstractmethod
-    def __init__(self):    
-        super(LearningRateSchedule, self).__init__()            
-
-    @abstractmethod
-    def _adjust_learning_rate(self, batch, logs):
-        pass
+    def __init__(self, min_lr=1e-10):    
+        super(LearningRateSchedule, self).__init__()    
+        self.min_lr = min_lr        
 
     def on_train_begin(self, logs=None):
         super(LearningRateSchedule, self).on_train_begin(logs)        
         self._eta0 = self.model.learning_rate
+        self._learning_rate_function = get_learning_rate_function(self)
+        self._attributes = vars()
 
     def on_batch_end(self, batch, logs=None):
         super(LearningRateSchedule, self).on_batch_end(batch=batch, logs=logs)        
-        self._adjust_learning_rate(batch=batch, logs=logs)
+        self.model.eta = max(self._learning_rate_function(attributes=self._attributes, 
+                                     eta0=self._eta0,
+                                     batch=batch), self.min_lr)
 
 # --------------------------------------------------------------------------  #
 class Constant(LearningRateSchedule):        
@@ -62,11 +343,8 @@ class Constant(LearningRateSchedule):
         self.name = "Constant Learning Rate"        
         super(Constant, self).__init__()
 
-    def _adjust_learning_rate(self, batch, logs):        
-        self.model.eta = self._eta0
-
 # --------------------------------------------------------------------------  #
-class LearningRateAnnealer(LearningRateSchedule):
+class LearningRateSchedule(ABC, LearningRateSchedule):
     """Base class for learning rate schedules. 
     
     Parameters
@@ -77,7 +355,7 @@ class LearningRateAnnealer(LearningRateSchedule):
 
     @abstractmethod
     def __init__(self, decay_factor=1):    
-        super(LearningRateAnnealer, self).__init__()
+        super(LearningRateSchedule, self).__init__()
         self.decay_factor = decay_factor        
 
     def _default_decay_factor(self):
@@ -96,13 +374,8 @@ class LearningRateAnnealer(LearningRateSchedule):
         """Performs validation of hyperparameters."""
         validate_zero_to_one(self.decay_factor, 'decay_factor')
 
-    @abstractmethod
-    def _adjust_learning_rate(self, batch, logs):
-        pass
-
     def on_train_begin(self, logs=None):
-        super(LearningRateAnnealer, self).on_train_begin(logs)        
-        self._eta0 = self.model.learning_rate
+        super(LearningRateSchedule, self).on_train_begin(logs)        
         if self.decay_factor is 'optimal':
             self.decay_factor = self._default_decay_factor()
         else:
@@ -110,36 +383,31 @@ class LearningRateAnnealer(LearningRateSchedule):
  
 
 # --------------------------------------------------------------------------  #
-class StepDecay(LearningRateAnnealer):
-    """ Time decay learning rate schedule as:
+class StepDecay(LearningRateSchedule):
+    """ Step decay learning rate schedule as:
     .. math:: \eta_0 \times \gamma^{\text{floor((1+batch)/decay_steps)}}
+
     Parameters
     ----------
-    decay_factor : float (default=1) or 'optimal'
-        If 'optimal', the decay rate will be computed based upon the 
-        learning rate and the anticipated number of batches
-    decay_steps : int (default=1)
+    decay_factor : float (default=0.5)
+        The factor by which the learning rate is reduced 
+    
+    decay_steps : int (default=10)
         The number of steps between each update
 
     """
 
-    def __init__(self, decay_factor=1, decay_steps=1):        
-        super(StepDecay, self).__init__(decay_factor=decay_factor)              
+    def __init__(self, decay_factor=0.5, decay_steps=10):        
+        super(StepDecay, self).__init__()              
         self.name = "Step Decay Learning Rate Schedule"
-        self.decay_steps = decay_steps
-
-    def _validate(self):
-        """Performs hyperparameter validation """
-        super(StepDecay, self)._validate()
-        validate_int(param=self.decay_steps, param_name='decay_steps', minimum=1)
-
-    def _adjust_learning_rate(self, batch, logs):
-        self.model.eta = self._eta0 * np.power(self.decay_factor, math.floor((1+batch)/self.decay_steps))
+        self.decay_factor = decay_factor
+        self.decay_steps = decay_steps       
 
 # --------------------------------------------------------------------------  #
-class TimeDecay(LearningRateAnnealer):
+class TimeDecay(LearningRateSchedule):
     """ Time decay learning rate schedule as:
     .. math:: \eta_t=\frac{\eta_0}{1+b\cdot t} 
+    
     Parameters
     ----------
     decay_factor : float (default=1) or 'optimal'
@@ -149,15 +417,12 @@ class TimeDecay(LearningRateAnnealer):
     """
 
     def __init__(self, decay_factor=.005):        
-        super(TimeDecay, self).__init__(
-            decay_factor=decay_factor)
+        super(TimeDecay, self).__init__()
         self.name = "Time Decay Learning Rate Schedule"              
-
-    def _adjust_learning_rate(self, batch, logs):
-        self.model.eta = self._eta0 / (1 + self.decay_factor * batch)
+        self.decay_factor = decay_factor
 
 # --------------------------------------------------------------------------  #
-class SqrtTimeDecay(LearningRateAnnealer):
+class SqrtTimeDecay(LearningRateSchedule):
     """ Time decay learning rate schedule as:
     .. math:: \eta_t=\frac{\eta_0}{1+b\cdot \sqrt{t}} 
     Parameters
@@ -172,12 +437,8 @@ class SqrtTimeDecay(LearningRateAnnealer):
             decay_factor=decay_factor)              
         self.name = "Sqrt Time Decay Learning Rate Schedule"
 
-    def _adjust_learning_rate(self, batch, logs):
-        self.model.eta = self._eta0 / (1 + self.decay_factor * \
-            np.sqrt(batch))        
-
 # --------------------------------------------------------------------------  #
-class ExponentialDecay(LearningRateAnnealer):
+class ExponentialDecay(LearningRateSchedule):
     """ Exponential decay learning rate schedule as:
     .. math:: \eta_t=\eta_0 \cdot \text{exp}(-b\cdot t)
     Parameters
@@ -192,17 +453,15 @@ class ExponentialDecay(LearningRateAnnealer):
             decay_factor=decay_factor)   
         self.name = "Exponential Decay Learning Rate Schedule"
 
-    def _adjust_learning_rate(self, batch, logs):
-        self.model.eta = self._eta0 * np.exp(-self.decay_factor * batch)
-
 # --------------------------------------------------------------------------  #
-class PolynomialDecay(LearningRateAnnealer):
+class PolynomialDecay(LearningRateSchedule):
     """ Polynomial decay learning rate schedule as:
     .. math:: \eta_t=\eta_0 \cdot \text{exp}(-b\cdot t)
+
     Parameters
     ----------
     power : float (default=1)
-        The power to which 
+        The power by which the learning rate will be annealed. 
 
     """
 
@@ -221,31 +480,10 @@ class PolynomialDecay(LearningRateAnnealer):
         batches = math.ceil(m / batch_size) * self.model.epochs
         decay = (1 - (batch / float(batches))) ** self.power                
         self.model.eta = self._eta0 * decay
-# --------------------------------------------------------------------------  #
-class ExponentialLearningRate(LearningRateAnnealer):
-    """ Exponential learning rate schedule as:
-    .. math:: \eta_t=(1=\lambda\eta)^{-2t-1}\eta
-    Reference : https://arxiv.org/abs/1910.07454
-    Parameters
-    ----------
-    decay_factor : float (default=1) or 'optimal'
-        If 'optimal', the decay rate will be computed based upon the 
-        learning rate and the anticipated number of batches
-    freq : str
-        The unit of time associated with a single batch. 
-    """
 
-    def __init__(self, decay_factor=.005):   
-        super(ExponentialLearningRate, self).__init__(
-            decay_factor=decay_factor)
-        self.name = "Exponential Learning Rate Schedule"
-
-    def _adjust_learning_rate(self, batch, logs):
-        self.model.eta = np.power((1- self.decay_factor*self._eta0), \
-            (-2*batch-1)) * self._eta0
 # --------------------------------------------------------------------------  #
-class ExponentialSchedule(LearningRateAnnealer):
-    """ Exponential decay learning rate schedule as:
+class ExponentialStepDecay(LearningRateSchedule):
+    """ Exponential step decay learning rate schedule as:
     .. math:: \eta_t=\eta_0 \cdot 10^{\frac{-t}{r}}
     Parameters
     ----------
@@ -256,13 +494,13 @@ class ExponentialSchedule(LearningRateAnnealer):
     """
 
     def __init__(self, decay_factor=0.0005, decay_steps=100):   
-        super(ExponentialSchedule, self).__init__(decay_factor=decay_factor)    
+        super(ExponentialStepDecay, self).__init__(decay_factor=decay_factor)    
         self.name = "Exponential Schedule Learning Rate Schedule"
         self.decay_steps = decay_steps
 
     def _validate(self):
         """Performs hyperparameter """
-        super(ExponentialSchedule, self)._validate()
+        super(ExponentialStepDecay, self)._validate()
         validate_int(param=self.decay_steps, param_name='decay_steps')
 
     def _adjust_learning_rate(self, batch, logs):
@@ -270,12 +508,12 @@ class ExponentialSchedule(LearningRateAnnealer):
             (batch / self.decay_steps))
 
 # --------------------------------------------------------------------------  #
-class PowerSchedule(LearningRateAnnealer):
-    """ Exponential decay learning rate schedule as:
+class PowerSchedule(LearningRateSchedule):
+    """ Power learning rate schedule as:
     .. math:: \eta_t=\eta_0 / (1+\frac{t}{r})^{c}
     Parameters
     ----------
-    power : float (default=1)
+    power : float (default=1.0)
         The factor by which the learning rate is decayed
     decay_steps : int
         The number of steps between each update
@@ -296,12 +534,13 @@ class PowerSchedule(LearningRateAnnealer):
         self.model.eta = self._eta0 / (1 + batch/self.decay_steps)**self.power
 
 # --------------------------------------------------------------------------  #
-class BottouSchedule(LearningRateAnnealer):
+class BottouSchedule(LearningRateSchedule):
     """ Learning rate schedule as described in:
     https://cilvr.cs.nyu.edu/diglib/lsml/bottou-sgd-tricks-2012.pdf
+    
     Parameters
     ----------
-    alpha : float (default=0.01)
+    decay_factor : float (default=0.01)
         The factor by which the learning rate is decayed
     """
 
@@ -316,7 +555,7 @@ class BottouSchedule(LearningRateAnnealer):
 # --------------------------------------------------------------------------- #
 #                             STABILITY                                       #
 # --------------------------------------------------------------------------- #
-class Improvement(LearningRateAnnealer):
+class Improvement(LearningRateSchedule):
     """Decays the learning rate if performance plateaus.
     Parameters
     ----------
